from enum import IntEnum


class Alignment(IntEnum):
    """Enum for text alignment types within a table cell

    A list of alignment types can be used to align each column individually::

        from table2ascii import Alignment, table2ascii

        table2ascii(
            header=["Product", "Category", "Price", "Rating"],
            body=[
                ["Milk", "Dairy", "$2.99", "6.28318"],
                ["Cheese", "Dairy", "$10.99", "8.2"],
                ["Apples", "Produce", "$0.99", "10.00"],
            ],
            # Align the first column to the left, the second to the center,
            # the third to the right, and the fourth to the decimal point
            alignments=[Alignment.LEFT, Alignment.CENTER, Alignment.RIGHT, Alignment.DECIMAL],
        )

        \"\"\"
        ╔════════════════════════════════════════╗
        ║ Product   Category    Price    Rating  ║
        ╟────────────────────────────────────────╢
        ║ Milk       Dairy      $2.99    6.28318 ║
        ║ Cheese     Dairy     $10.99    8.2     ║
        ║ Apples    Produce     $0.99   10.00    ║
        ╚════════════════════════════════════════╝
        \"\"\"

    A single alignment type can be used to align all columns::

        table2ascii(
            header=["First Name", "Last Name", "Age"],
            body=[
                ["John", "Smith", 30],
                ["Jane", "Doe", 28],
            ],
            alignments=Alignment.LEFT,  # Align all columns to the left
            number_alignments=Alignment.RIGHT,  # Align all numeric values to the right
        )

        \"\"\"
        ╔══════════════════════════════╗
        ║ First Name   Last Name   Age ║
        ╟──────────────────────────────╢
        ║ John         Smith        30 ║
        ║ Jane         Doe          28 ║
        ╚══════════════════════════════╝
        \"\"\"

    .. note::

        If :attr:`DECIMAL` is used in the ``number_alignments`` argument to :func:`table2ascii`,
        all non-numeric values will be aligned according to the ``alignments`` argument.
        If the :attr:`DECIMAL` alignment type is used in the ``alignments`` argument,
        all non-numeric values will be aligned to the center.
        Numeric values include integers, floats, and strings containing only :meth:`decimal <str.isdecimal>`
        characters and at most one decimal point.

    .. versionchanged:: 1.1.0

        Added :attr:`DECIMAL` alignment -- align decimal numbers such that
        the decimal point is aligned with the decimal point of all other numbers
        in the same column.
    """

    LEFT = 0
    CENTER = 1
    RIGHT = 2
    DECIMAL = 3



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

from .alignment import Alignment
from .annotations import SupportsStr


class Table2AsciiError(Exception):
    """Base class for all table2ascii exceptions"""

    def _message(self) -> str:
        """Return the error message"""
        raise NotImplementedError


class TableOptionError(Table2AsciiError, ValueError):
    """Base class for exceptions raised when an invalid option
    is passed when creating an ascii table

    This class is a subclass of :class:`Table2AsciiError` and :class:`ValueError`.
    """


class ColumnCountMismatchError(TableOptionError):
    """Base class for exceptions raised when a parameter has an
    invalid number of columns

    This class is a subclass of :class:`TableOptionError`.
    """

    expected_columns: int


class FooterColumnCountMismatchError(ColumnCountMismatchError):
    """Exception raised when the number of columns in the footer
    does not match the number of columns in the header

    This class is a subclass of :class:`ColumnCountMismatchError`.

    Attributes:
        footer (:class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]):
            The footer that caused the error
        expected_columns (:class:`int`): The number of columns that were expected
    """

    def __init__(self, footer: Sequence[SupportsStr], expected_columns: int):
        self.footer = footer
        self.expected_columns = expected_columns
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Footer column count mismatch: {len(self.footer)} columns "
            f"found, expected {self.expected_columns}."
        )


class BodyColumnCountMismatchError(ColumnCountMismatchError):
    """Exception raised when the number of columns in the body
    does not match the number of columns in the footer or header

    This class is a subclass of :class:`ColumnCountMismatchError`.

    Attributes:
        body (:class:`Sequence <collections.abc.Sequence>`\ [\ :class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]]):
            The body that caused the error
        expected_columns (:class:`int`): The number of columns that were expected
        first_invalid_row (:class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]):
            The first row with an invalid column count
    """

    def __init__(self, body: Sequence[Sequence[SupportsStr]], expected_columns: int):
        self.body = body
        self.expected_columns = expected_columns
        self.first_invalid_row = next(
            (row for row in self.body if len(row) != self.expected_columns)
        )
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Body column count mismatch: A row with {len(self.first_invalid_row)} "
            f"columns was found, expected {self.expected_columns}."
        )


class AlignmentCountMismatchError(ColumnCountMismatchError):
    """Exception raised when the number of alignments does not match
    the number of columns in the table

    This class is a subclass of :class:`ColumnCountMismatchError`.

    Attributes:
        alignments (:class:`Sequence <collections.abc.Sequence>`\ [:class:`Alignment`]):
            The alignments that caused the error
        expected_columns (:class:`int`): The number of columns that were expected
    """

    def __init__(self, alignments: Sequence[Alignment], expected_columns: int):
        self.alignments = alignments
        self.expected_columns = expected_columns
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Alignment count mismatch: {len(self.alignments)} alignments "
            f"found, expected {self.expected_columns}."
        )


class ColumnWidthsCountMismatchError(ColumnCountMismatchError):
    """Exception raised when the number of column widths does not match
    the number of columns in the table

    This class is a subclass of :class:`ColumnCountMismatchError`.

    Attributes:
        column_widths (:class:`Sequence <collections.abc.Sequence>`\ [:data:`Optional <typing.Optional>`\ [:class:`int`]]):
            The column widths that caused the error
        expected_columns (:class:`int`): The number of columns that were expected
    """

    def __init__(self, column_widths: Sequence[int | None], expected_columns: int):
        self.column_widths = column_widths
        self.expected_columns = expected_columns
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Column widths count mismatch: {len(self.column_widths)} column widths "
            f"found, expected {self.expected_columns}."
        )


class NoHeaderBodyOrFooterError(TableOptionError):
    """Exception raised when no header, body or footer is provided

    This class is a subclass of :class:`TableOptionError`.
    """

    def __init__(self):
        super().__init__(self._message())

    def _message(self) -> str:
        return "At least one of header, body or footer must be provided."


class InvalidCellPaddingError(TableOptionError):
    """Exception raised when the cell padding is invalid

    This class is a subclass of :class:`TableOptionError`.

    Attributes:
        padding (:class:`int`): The padding that caused the error
    """

    def __init__(self, padding: int):
        self.padding = padding
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Invalid cell padding: The cell padding provided was {self.padding} "
            f"but it must be a non-negative integer."
        )


class ColumnWidthTooSmallError(TableOptionError):
    """Exception raised when the column width is smaller than the minimum
    number of characters that are required to display the content

    This class is a subclass of :class:`TableOptionError`.

    Attributes:
        column_index (:class:`int`): The index of the column that caused the error
        column_width (:class:`int`): The column width that caused the error
        min_width (:class:`int`): The minimum width that is allowed
    """

    def __init__(self, column_index: int, column_width: int, min_width: int | None = None):
        self.column_index = column_index
        self.column_width = column_width
        self.min_width = min_width
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Column width too small: The column width for index {self.column_index} "
            f"of `column_widths` is {self.column_width}, but the minimum width "
            f"required to display the content is {self.min_width}."
        )


class InvalidColumnWidthError(ColumnWidthTooSmallError):
    """Exception raised when the column width is invalid

    This class is a subclass of :class:`ColumnWidthTooSmallError`.
    """

    def _message(self) -> str:
        return (
            f"Invalid column width: The column width for index {self.column_index} "
            f"of `column_widths` is {self.column_width}, but the column width "
            f"must be a positive integer."
        )


class InvalidAlignmentError(TableOptionError):
    """Exception raised when an invalid value is passed for an :class:`Alignment`

    This class is a subclass of :class:`TableOptionError`.

    Attributes:
        alignment (:data:`Any <typing.Any>`): The alignment value that caused the error
    """

    def __init__(self, alignment: Any):
        self.alignment = alignment
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Invalid alignment: {self.alignment!r} is not a valid alignment. "
            f"Valid alignments are: {', '.join(a.__repr__() for a in Alignment)}"
        )


class TableStyleTooLongError(Table2AsciiError, ValueError):
    """Exception raised when the number of characters passed in the string
    for creating the table style exceeds the number of parameters that the
    table style accepts

    This class is a subclass of :class:`Table2AsciiError` and :class:`ValueError`.

    Attributes:
        string (:class:`str`): The string that caused the error
        max_characters (:class:`int`): The maximum number of characters that are allowed
    """

    def __init__(self, string: str, max_characters: int):
        self.string = string
        self.max_characters = max_characters
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Too many characters for table style: {len(self.string)} characters "
            f"found, but the maximum number of characters allowed is {self.max_characters}."
        )


class TableStyleTooShortWarning(UserWarning):
    """Warning raised when the number of characters passed in the string
    for creating the table style is fewer than the number of parameters
    that the table style accepts

    This class is a subclass of :class:`UserWarning`.

    It can be silenced using :func:`warnings.filterwarnings`.

    Attributes:
        string (:class:`str`): The string that caused the warning
        max_characters (:class:`int`): The number of characters that :class:`TableStyle` accepts
    """

    def __init__(self, string: str, max_characters: int):
        self.string = string
        self.max_characters = max_characters
        super().__init__(self._message())

    def _message(self) -> str:
        return (
            f"Too few characters for table style: {len(self.string)} characters "
            f"found, but table styles can accept {self.max_characters} characters. "
            f"Missing characters will be replaced with spaces."
        )



from enum import Enum


class Merge(Enum):
    """Enum for merging table cells

    Using :attr:`Merge.LEFT` in a table cell will merge the cell it is used in
    with the cell to its left.

    In the case that the contents of the merged cell are longer than the
    combined widths of the unmerged cells in the rows above and below,
    the merged cell will be wrapped onto multiple lines. The ``column_widths``
    option can be used to control the widths of the unmerged cells.

    Example::

        from table2ascii import Merge, PresetStyle, table2ascii

        table2ascii(
            header=["Name", "Price", "Category", "Stock"],
            body=[["Milk", "$2.99", "N/A", Merge.LEFT]],
            footer=["Description", "Milk is a nutritious beverage", Merge.LEFT, Merge.LEFT],
            style=PresetStyle.double_box,
        )

        \"\"\"
        ╔═════════════╦═══════╦══════════╦═══════╗
        ║    Name     ║ Price ║ Category ║ Stock ║
        ╠═════════════╬═══════╬══════════╩═══════╣
        ║    Milk     ║ $2.99 ║       N/A        ║
        ╠═════════════╬═══════╩══════════════════╣
        ║ Description ║   Milk is a nutritious   ║
        ║             ║         beverage         ║
        ╚═════════════╩══════════════════════════╝
        \"\"\"

    .. versionadded:: 1.0.0
    """

    LEFT = 0

    def __str__(self) -> str:
        return ""



"""
table2ascii - Library for converting 2D Python lists to fancy ASCII/Unicode tables
"""
import sys
from typing import TYPE_CHECKING

from .alignment import Alignment
from .annotations import SupportsStr
from .exceptions import (
    AlignmentCountMismatchError,
    BodyColumnCountMismatchError,
    ColumnCountMismatchError,
    ColumnWidthsCountMismatchError,
    ColumnWidthTooSmallError,
    FooterColumnCountMismatchError,
    InvalidAlignmentError,
    InvalidCellPaddingError,
    InvalidColumnWidthError,
    Table2AsciiError,
    TableOptionError,
    TableStyleTooLongError,
    TableStyleTooShortWarning,
)
from .merge import Merge
from .preset_style import PresetStyle
from .table_style import TableStyle
from .table_to_ascii import table2ascii

if TYPE_CHECKING or sys.version_info >= (3, 8):
    from importlib import metadata
else:
    import importlib_metadata as metadata

__version__ = metadata.version(__name__)

__all__ = [
    "Alignment",
    "Merge",
    "PresetStyle",
    "TableStyle",
    "table2ascii",
    "AlignmentCountMismatchError",
    "BodyColumnCountMismatchError",
    "ColumnCountMismatchError",
    "ColumnWidthsCountMismatchError",
    "ColumnWidthTooSmallError",
    "FooterColumnCountMismatchError",
    "InvalidAlignmentError",
    "InvalidCellPaddingError",
    "InvalidColumnWidthError",
    "Table2AsciiError",
    "TableOptionError",
    "TableStyleTooLongError",
    "TableStyleTooShortWarning",
    "SupportsStr",
]



from dataclasses import dataclass
import warnings

from .exceptions import TableStyleTooShortWarning, TableStyleTooLongError


@dataclass
class TableStyle:
    """Class for storing information about a table style

    Parts of the table labeled alphabetically from A to V:

    .. code-block::

        ABBBBBCBBBBBDBBBBBDBBBBBDBBBBBE
        F     G     H     H     H     F
        IJJJJJKJJJJJLJJJJJLJJJJJLJJJJJM
        F     G     H     H     H     F
        NOOOOOPOOOOOQOOOOOQOOOOOQOOOOOR
        F     G     H     H     H     F
        IJJJJJKJJJJJLJJJJJLJJJJJLJJJJJM
        F     G     H     H     H     F
        SBBBBBTBBBBBUBBBBBUBBBBBUBBBBBV

    How the theme is displayed using :ref:`PresetStyle.double_thin_box`
    as an example:

    .. code-block::

        ╔═════╦═════╤═════╤═════╤═════╗
        ║  #  ║  G  │  H  │  R  │  S  ║
        ╠═════╬═════╪═════╪═════╪═════╣
        ║  1  ║ 30  │ 40  │ 35  │ 30  ║
        ╟─────╫─────┼─────┼─────┼─────╢
        ║  2  ║ 30  │ 40  │ 35  │ 30  ║
        ╠═════╬═════╪═════╪═════╪═════╣
        ║ SUM ║ 130 │ 140 │ 135 │ 130 ║
        ╚═════╩═════╧═════╧═════╧═════╝

    In addition to the parts above, W-Z and the four fields that follow
    (labeled 0-3) are used for top and bottom edges of merged cells as shown:

    .. code-block::

        ╔══════════════╤══════╤══════╗
        ║    Header    │      │      ║
        ╠══════[2]═════╪═════[Z]═════╣
        ║       ║      │             ║
        ╟──────[1]─────┼─────────────╢
        ║              │             ║
        ╟──────[0]─────┼─────[W]─────╢
        ║       ║      │      │      ║
        ╟───────╫──────┼─────[X]─────╢
        ║       ║      │             ║
        ╠══════[3]═════╪═════[Y]═════╣
        ║    Footer    │      │      ║
        ╚══════════════╧══════╧══════╝

        [W] = ┬ [X] = ┴ [Y] = ╤ [Z] = ╧
        [0] = ╥ [1] = ╨ [2] = ╦ [3] = ╩

    .. versionchanged:: 1.0.0

        Added fields for edges of merged cells:

        ``col_row_top_tee``, ``col_row_bottom_tee``, ``heading_row_top_tee``,
        ``heading_row_bottom_tee``, ``heading_col_body_row_top_tee``,
        ``heading_col_body_row_bottom_tee``, ``heading_col_heading_row_top_tee``,
        ``heading_col_heading_row_bottom_tee``
    """

    # parts of the table
    top_left_corner: str  # A
    top_and_bottom_edge: str  # B
    heading_col_top_tee: str  # C
    top_tee: str  # D
    top_right_corner: str  # E
    left_and_right_edge: str  # F
    heading_col_sep: str  # G
    col_sep: str  # H
    heading_row_left_tee: str  # I
    heading_row_sep: str  # J
    heading_col_heading_row_cross: str  # K
    heading_row_cross: str  # L
    heading_row_right_tee: str  # M
    row_left_tee: str  # N
    row_sep: str  # O
    heading_col_row_cross: str  # P
    col_row_cross: str  # Q
    row_right_tee: str  # R
    bottom_left_corner: str  # S
    heading_col_bottom_tee: str  # T
    bottom_tee: str  # U
    bottom_right_corner: str  # V
    col_row_top_tee: str  # W
    col_row_bottom_tee: str  # X
    heading_row_top_tee: str  # Y
    heading_row_bottom_tee: str  # Z
    heading_col_body_row_top_tee: str  # 0
    heading_col_body_row_bottom_tee: str  # 1
    heading_col_heading_row_top_tee: str  # 2
    heading_col_heading_row_bottom_tee: str  # 3

    @classmethod
    def from_string(cls, string: str) -> "TableStyle":
        """Create a TableStyle from a string

        .. versionchanged:: 1.0.0

            The string will be padded on the right with spaces if it is too short
            and :class:`~table2ascii.exceptions.TableStyleTooLongError` will be
            raised if it is too long.

        Args:
            string: The string to create the TableStyle from

        Returns:
            A TableStyle object

        Example::

            TableStyle.from_string("╔═╦╤╗║║│╠═╬╪╣╟─╫┼╢╚╩╧╝┬┴╤╧╥╨╦╩")

        Raises:
            TableStyleTooLongError: If the string is too long
        """
        num_params = len(cls.__dataclass_fields__)
        # if the string is too long, raise an error
        if len(string) > num_params:
            raise TableStyleTooLongError(string, num_params)
        # if the string is too short, show a warning and pad it with spaces
        elif len(string) < num_params:
            warnings.warn(TableStyleTooShortWarning(string, num_params), stacklevel=2)
            string += " " * (num_params - len(string))
        return cls(*string)

    def set(self, **kwargs: str) -> "TableStyle":
        """Set attributes of the TableStyle

        Args:
            kwargs: The attributes to set

        Returns:
            A TableStyle object with the attributes set

        Example::

            TableStyle.from_string("~" * 30).set(left_and_right_edge="", col_sep="")
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass

from .alignment import Alignment
from .table_style import TableStyle


@dataclass
class Options:
    """Class for storing options that the user sets

    .. versionchanged:: 1.1.0

        Added ``number_alignments`` option

    .. versionchanged:: 1.0.0

        Added ``use_wcwidth`` option
    """

    first_col_heading: bool
    last_col_heading: bool
    column_widths: Sequence[int | None] | None
    alignments: Sequence[Alignment] | Alignment | None
    number_alignments: Sequence[Alignment] | Alignment | None
    cell_padding: int
    style: TableStyle
    use_wcwidth: bool



from __future__ import annotations

import textwrap
from math import ceil, floor
from collections.abc import Sequence

from wcwidth import wcswidth

from .alignment import Alignment
from .annotations import SupportsStr
from .exceptions import (
    AlignmentCountMismatchError,
    BodyColumnCountMismatchError,
    ColumnWidthTooSmallError,
    ColumnWidthsCountMismatchError,
    FooterColumnCountMismatchError,
    InvalidAlignmentError,
    InvalidCellPaddingError,
    InvalidColumnWidthError,
    NoHeaderBodyOrFooterError,
)
from .merge import Merge
from .options import Options
from .preset_style import PresetStyle
from .table_style import TableStyle


class TableToAscii:
    """Class used to convert a 2D Python table to ASCII text"""

    def __init__(
        self,
        header: Sequence[SupportsStr] | None,
        body: Sequence[Sequence[SupportsStr]] | None,
        footer: Sequence[SupportsStr] | None,
        options: Options,
    ):
        """Validate arguments and initialize fields

        Args:
            header: The values in the header of the table
            body: The rows of values in the body of the table
            footer: The values in the footer of the table
            options: The options for the table
        """
        # initialize fields
        self.__header = list(header) if header else None
        self.__body = list([list(row) for row in body]) if body else None
        self.__footer = list(footer) if footer else None
        self.__style = options.style
        self.__first_col_heading = options.first_col_heading
        self.__last_col_heading = options.last_col_heading
        self.__cell_padding = options.cell_padding
        self.__use_wcwidth = options.use_wcwidth

        # calculate number of columns
        self.__columns = self.__count_columns()

        # check if footer has a different number of columns
        if footer and len(footer) != self.__columns:
            raise FooterColumnCountMismatchError(footer, self.__columns)
        # check if any rows in body have a different number of columns
        if body and any(len(row) != self.__columns for row in body):
            raise BodyColumnCountMismatchError(body, self.__columns)

        # check that at least one of header, body, or footer is not None
        if not header and not body and not footer:
            raise NoHeaderBodyOrFooterError()

        self.__alignments = self.__determine_alignments(
            options.alignments, default=Alignment.CENTER
        )
        self.__number_alignments = self.__determine_alignments(
            options.number_alignments, default=self.__alignments
        )

        # keep track of the number widths and positions of the decimal points for decimal alignment
        decimal_widths, decimal_positions = self.__calculate_decimal_widths_and_positions()
        self.__decimal_widths: list[int] = decimal_widths
        self.__decimal_positions: list[int] = decimal_positions

        # calculate or use given column widths
        self.__column_widths = self.__calculate_column_widths(options.column_widths)

        # check if the cell padding is valid
        if self.__cell_padding < 0:
            raise InvalidCellPaddingError(self.__cell_padding)

        # if any row starts with Merge.LEFT, replace it with an empty string
        self.__fix_rows_beginning_with_merge()

    def __count_columns(self) -> int:
        """Get the number of columns in the table based on the provided header, footer, and body lists.

        Returns:
            The number of columns in the table
        """
        if self.__header:
            return len(self.__header)
        if self.__footer:
            return len(self.__footer)
        if self.__body and len(self.__body) > 0:
            return len(self.__body[0])
        return 0

    def __determine_alignments(
        self,
        user_alignments: Sequence[Alignment] | Alignment | None,
        *,
        default: Sequence[Alignment] | Alignment,
    ) -> list[Alignment]:
        """Determine the alignments for each column based on the user provided alignments option.

        Args:
            user_alignments: The alignments specified by the user
            default: The default alignments to use if user_alignments is None

        Returns:
            The alignments for each column in the table
        """
        alignments = user_alignments if user_alignments is not None else default

        # if alignments is a single Alignment, convert it to a list of that Alignment
        if isinstance(alignments, Alignment):
            alignments = [alignments] * self.__columns

        # check if alignments specified have a different number of columns
        if len(alignments) != self.__columns:
            raise AlignmentCountMismatchError(alignments, self.__columns)

        return list(alignments)

    def __auto_column_widths(self) -> list[int]:
        """Get the minimum number of characters needed for the values in each column in the table
        with 1 space of padding on each side.

        Returns:
            The minimum number of characters needed for each column
        """

        def widest_line(value: SupportsStr) -> int:
            """Returns the width of the longest line in a multi-line string"""
            text = str(value)
            return max(self.__str_width(line) for line in text.splitlines()) if len(text) else 0

        def get_column_width(row: Sequence[SupportsStr], column: int) -> int:
            """Get the width of a cell in a column"""
            value = row[column]
            next_value = row[column + 1] if column < self.__columns - 1 else None
            if value is Merge.LEFT or next_value is Merge.LEFT:
                return 0
            return widest_line(value)

        column_widths = []
        # get the width necessary for each column
        for i in range(self.__columns):
            # number of characters in column of i of header, each body row, and footer
            header_size = get_column_width(self.__header, i) if self.__header else 0
            body_size = max(get_column_width(row, i) for row in self.__body) if self.__body else 0
            footer_size = get_column_width(self.__footer, i) if self.__footer else 0
            min_text_width = max(header_size, body_size, footer_size, self.__decimal_widths[i])
            # get the max and add 2 for padding each side with a space depending on cell padding
            column_widths.append(min_text_width + self.__cell_padding * 2)
        return column_widths

    def __calculate_decimal_widths_and_positions(self) -> tuple[list[int], list[int]]:
        """Calculate the positions of the decimal points for decimal alignment.

        Returns:
            A tuple of the widths of the decimal numbers in each column
            and the positions of the decimal points in each column
        """
        decimal_widths: list[int] = [0] * self.__columns
        decimal_positions: list[int] = [0] * self.__columns
        for i in range(self.__columns):
            # skip if the column is not decimal aligned
            if self.__number_alignments[i] != Alignment.DECIMAL:
                continue
            # list all values in the i-th column of header, body, and footer
            values = [str(self.__header[i])] if self.__header else []
            values += [str(row[i]) for row in self.__body] if self.__body else []
            values += [str(self.__footer[i])] if self.__footer else []
            # filter out values that are not numbers and split at the decimal point
            split_values = [
                self.__split_decimal(value) for value in values if self.__is_number(value)
            ]
            # skip if there are no decimal values
            if len(split_values) == 0:
                continue
            # get the max number of digits before and after the decimal point
            max_before_decimal = max(self.__str_width(parts[0]) for parts in split_values)
            max_after_decimal = max(self.__str_width(parts[1]) for parts in split_values)
            # add 1 for the decimal point if there are any decimal point values
            has_decimal = any(self.__is_number(value) and "." in value for value in values)
            # store the total width of the decimal numbers in the column
            decimal_widths[i] = max_before_decimal + max_after_decimal + int(has_decimal)
            # store the max digits before the decimal point for decimal alignment
            decimal_positions[i] = max_before_decimal
        return decimal_widths, decimal_positions

    def __calculate_column_widths(
        self, user_column_widths: Sequence[int | None] | None
    ) -> list[int]:
        """Calculate the width of each column in the table based on the cell values and provided column widths.

        Args:
            user_column_widths: The user specified column widths

        Returns:
            The width of each column in the table
        """
        column_widths = self.__auto_column_widths()
        if user_column_widths:
            # check that the right number of columns were specified
            if len(user_column_widths) != self.__columns:
                raise ColumnWidthsCountMismatchError(user_column_widths, self.__columns)
            # check that each column is at least as large as the minimum size
            for i in range(len(user_column_widths)):
                option = user_column_widths[i]
                minimum = column_widths[i]
                if option is None:
                    option = minimum
                elif option < 0:
                    raise InvalidColumnWidthError(i, option)
                elif option < minimum:
                    raise ColumnWidthTooSmallError(i, option, minimum)
                column_widths[i] = option
        return column_widths

    def __fix_rows_beginning_with_merge(self) -> None:
        """Fix rows that begin with Merge.LEFT by replacing the cell with an empty string."""
        if self.__body:
            for row_index, row in enumerate(self.__body):
                if row and row[0] == Merge.LEFT:
                    self.__body[row_index][0] = ""
        if self.__header and self.__header[0] == Merge.LEFT:
            self.__header[0] = ""
        if self.__footer and self.__footer[0] == Merge.LEFT:
            self.__footer[0] = ""

    def __pad(self, cell_value: SupportsStr, width: int, col_index: int) -> str:
        """Pad a string of text to a given width with specified alignment

        Args:
            cell_value: The text in the cell to pad
            width: The width in characters to pad to
            col_index: The index of the column

        Returns:
            The padded text
        """
        alignment = self.__alignments[col_index]
        text = str(cell_value)
        # set alignment for numeric values
        if self.__is_number(text):
            # if the number alignment is decimal, pad such that the decimal point
            # is aligned to the column's decimal position and use the default alignment
            if self.__number_alignments[col_index] == Alignment.DECIMAL:
                decimal_position = self.__decimal_positions[col_index]
                decimal_max_width = self.__decimal_widths[col_index]
                text_before_decimal = self.__split_decimal(text)[0]
                before = " " * (decimal_position - self.__str_width(text_before_decimal))
                after = " " * (decimal_max_width - self.__str_width(text) - len(before))
                text = f"{before}{text}{after}"
            # otherwise use the number alignment as the alignment for the cell
            else:
                alignment = self.__number_alignments[col_index]
        # add minimum cell padding around the text
        padding = " " * self.__cell_padding
        padded_text = f"{padding}{text}{padding}"
        text_width = self.__str_width(padded_text)
        # pad the text based on the alignment
        if alignment == Alignment.LEFT:
            # pad with spaces on the end
            return padded_text + (" " * (width - text_width))
        elif alignment in (Alignment.CENTER, Alignment.DECIMAL):
            # pad with spaces, half on each side
            before = " " * floor((width - text_width) / 2)
            after = " " * ceil((width - text_width) / 2)
            return before + padded_text + after
        elif alignment == Alignment.RIGHT:
            # pad with spaces at the beginning
            return (" " * (width - text_width)) + padded_text
        raise InvalidAlignmentError(alignment)

    def __wrap_long_lines_in_merged_cells(
        self, row: Sequence[SupportsStr], column_separator: str
    ) -> list[SupportsStr]:
        """Wrap long lines in merged cells to the width of the merged cell

        Args:
            row: The row to wrap cells in
            column_separator: The column separator between cells

        Returns:
            The row with long lines wrapped
        """
        wrapped_row: list[SupportsStr] = []
        for col_index, cell in enumerate(row):
            if cell is Merge.LEFT:
                wrapped_row.append(cell)
                continue
            merged_width = self.__column_widths[col_index]
            # if the columns to the right are Merge.LEFT, add their width to the padding
            for other_col_index in range(col_index + 1, self.__columns):
                if row[other_col_index] is not Merge.LEFT:
                    break
                merged_width += self.__column_widths[other_col_index] + len(column_separator)
            cell = textwrap.fill(str(cell), merged_width - self.__cell_padding * 2)
            wrapped_row.append(cell)
        return wrapped_row

    def __row_to_ascii(
        self,
        left_edge: str,
        heading_col_sep: str,
        column_separator: str,
        right_edge: str,
        filler: str | Sequence[SupportsStr],
        previous_content_row: Sequence[SupportsStr] | None = None,
        next_content_row: Sequence[SupportsStr] | None = None,
        top_tee: str | None = None,
        bottom_tee: str | None = None,
        heading_col_top_tee: str | None = None,
        heading_col_bottom_tee: str | None = None,
    ) -> str:
        """Assembles a line of text in the ascii table

        Returns:
            The line in the ascii table
        """
        output = ""
        # wrap long lines in merged cells
        if isinstance(filler, list):
            filler = self.__wrap_long_lines_in_merged_cells(filler, column_separator)
        # find the maximum number of lines a single cell in the column has (minimum of 1)
        num_lines = max(len(str(cell).splitlines()) for cell in filler) or 1
        # repeat for each line of text in the cell
        for line_index in range(num_lines):
            output += self.__line_in_row_to_ascii(
                line_index=line_index,
                left_edge=left_edge,
                heading_col_sep=heading_col_sep,
                column_separator=column_separator,
                right_edge=right_edge,
                filler=filler,
                previous_content_row=previous_content_row,
                next_content_row=next_content_row,
                top_tee=top_tee,
                bottom_tee=bottom_tee,
                heading_col_top_tee=heading_col_top_tee,
                heading_col_bottom_tee=heading_col_bottom_tee,
            )
        # don't use separation row if it's only space
        if isinstance(filler, str) and output.strip() == "":
            output = ""
        return output

    def __line_in_row_to_ascii(
        self,
        line_index: int,
        left_edge: str,
        heading_col_sep: str,
        column_separator: str,
        right_edge: str,
        filler: str | Sequence[SupportsStr],
        previous_content_row: Sequence[SupportsStr] | None = None,
        next_content_row: Sequence[SupportsStr] | None = None,
        top_tee: str | None = None,
        bottom_tee: str | None = None,
        heading_col_top_tee: str | None = None,
        heading_col_bottom_tee: str | None = None,
    ) -> str:
        """Assembles a line of text in the ascii table

        Returns:
            The line in the ascii table
        """
        output = left_edge
        # add columns
        for col_index in range(self.__columns):
            output += self.__line_in_cell_column_to_ascii(
                line_index=line_index,
                col_index=col_index,
                heading_col_sep=heading_col_sep,
                column_separator=column_separator,
                right_edge=right_edge,
                filler=filler,
                previous_content_row=previous_content_row,
                next_content_row=next_content_row,
                top_tee=top_tee,
                bottom_tee=bottom_tee,
                heading_col_top_tee=heading_col_top_tee,
                heading_col_bottom_tee=heading_col_bottom_tee,
            )
        output += "\n"
        return output

    def __line_in_cell_column_to_ascii(
        self,
        line_index: int,
        col_index: int,
        heading_col_sep: str,
        column_separator: str,
        right_edge: str,
        filler: str | Sequence[SupportsStr],
        previous_content_row: Sequence[SupportsStr] | None = None,
        next_content_row: Sequence[SupportsStr] | None = None,
        top_tee: str | None = None,
        bottom_tee: str | None = None,
        heading_col_top_tee: str | None = None,
        heading_col_bottom_tee: str | None = None,
    ) -> str:
        """Assembles a column of text in the ascii table

        Returns:
            The column in the ascii table
        """
        output = ""
        col_content = (
            # if filler is a separator character, repeat it for the full width of the column
            filler * self.__column_widths[col_index]
            if isinstance(filler, str)
            # otherwise, use the text from the corresponding column in the filler list
            else self.__get_padded_cell_line_content(
                line_index=line_index,
                col_index=col_index,
                column_separator=column_separator,
                filler=filler,
            )
        )
        output += col_content
        # check for merged cells
        next_value = prev_row_next_value = next_row_next_value = None
        if col_index < self.__columns - 1:
            next_value = filler[col_index + 1] if isinstance(filler, list) else None
            prev_row_next_value = (
                previous_content_row[col_index + 1] if previous_content_row else None
            )
            next_row_next_value = next_content_row[col_index + 1] if next_content_row else None
        # column separator
        sep = column_separator
        # handle separators between rows when previous or next row is a merged cell
        if top_tee and prev_row_next_value is Merge.LEFT:
            sep = top_tee
        if bottom_tee and next_row_next_value is Merge.LEFT:
            sep = bottom_tee
        if (
            isinstance(filler, str)
            and prev_row_next_value in (Merge.LEFT, None)
            and next_row_next_value in (Merge.LEFT, None)
        ):
            sep = filler
        # use column heading if first or last column option is specified
        if (col_index == 0 and self.__first_col_heading) or (
            col_index == self.__columns - 2 and self.__last_col_heading
        ):
            sep = heading_col_sep
            # handle separators between rows when previous or next row is a merged cell
            if heading_col_top_tee and prev_row_next_value is Merge.LEFT:
                sep = heading_col_top_tee
            if heading_col_bottom_tee and next_row_next_value is Merge.LEFT:
                sep = heading_col_bottom_tee
        # replace last separator with symbol for edge of the row
        elif col_index == self.__columns - 1:
            sep = right_edge
        # if this is cell contents and the next column is Merge.LEFT, don't add a separator
        if next_value is Merge.LEFT:
            sep = ""
        return output + sep

    def __get_padded_cell_line_content(
        self, line_index: int, col_index: int, column_separator: str, filler: Sequence[SupportsStr]
    ) -> str:
        # If this is a merge cell, merge with the previous column
        if filler[col_index] is Merge.LEFT:
            return ""
        # get the text of the current line in the cell
        # if there are fewer lines in the current cell than others, empty string is used
        col_lines = str(filler[col_index]).splitlines()
        col_content = col_lines[line_index] if line_index < len(col_lines) else ""
        pad_width = self.__column_widths[col_index]
        # if the columns to the right are Merge.LEFT, add their width to the padding
        for other_col_index in range(col_index + 1, self.__columns):
            if filler[other_col_index] is not Merge.LEFT:
                break
            pad_width += self.__column_widths[other_col_index] + len(column_separator)
        # pad the text to the width of the column using the alignment
        return self.__pad(
            cell_value=col_content,
            width=pad_width,
            col_index=col_index,
        )

    def __top_edge_to_ascii(self) -> str:
        """Assembles the top edge of the ascii table

        Returns:
            The top edge of the ascii table
        """
        first_row = self.__body[0] if self.__body else None
        first_row = self.__header if self.__header else first_row
        return self.__row_to_ascii(
            left_edge=self.__style.top_left_corner,
            heading_col_sep=self.__style.heading_col_top_tee,
            column_separator=self.__style.top_tee,
            right_edge=self.__style.top_right_corner,
            filler=self.__style.top_and_bottom_edge,
            next_content_row=first_row,
            top_tee=self.__style.col_row_top_tee,
            bottom_tee=self.__style.top_and_bottom_edge,
            heading_col_top_tee=self.__style.heading_col_top_tee,
            heading_col_bottom_tee=self.__style.top_and_bottom_edge,
        )

    def __bottom_edge_to_ascii(self) -> str:
        """Assembles the bottom edge of the ascii table

        Returns:
            The bottom edge of the ascii table
        """
        last_row = self.__body[-1] if self.__body else None
        last_row = self.__footer if self.__footer else last_row
        return self.__row_to_ascii(
            left_edge=self.__style.bottom_left_corner,
            heading_col_sep=self.__style.heading_col_bottom_tee,
            column_separator=self.__style.bottom_tee,
            right_edge=self.__style.bottom_right_corner,
            filler=self.__style.top_and_bottom_edge,
            previous_content_row=last_row,
            top_tee=self.__style.top_and_bottom_edge,
            bottom_tee=self.__style.col_row_bottom_tee,
            heading_col_top_tee=self.__style.top_and_bottom_edge,
            heading_col_bottom_tee=self.__style.heading_col_bottom_tee,
        )

    def __content_row_to_ascii(self, row: Sequence[SupportsStr]) -> str:
        """Assembles a row of cell values into a single line of the ascii table

        Returns:
            The row of the ascii table
        """
        return self.__row_to_ascii(
            left_edge=self.__style.left_and_right_edge,
            heading_col_sep=self.__style.heading_col_sep,
            column_separator=self.__style.col_sep,
            right_edge=self.__style.left_and_right_edge,
            filler=row,
        )

    def __heading_sep_to_ascii(
        self,
        previous_content_row: Sequence[SupportsStr] | None = None,
        next_content_row: Sequence[SupportsStr] | None = None,
    ) -> str:
        """Assembles the separator below the header or above footer of the ascii table

        Returns:
            The separator line
        """
        return self.__row_to_ascii(
            left_edge=self.__style.heading_row_left_tee,
            heading_col_sep=self.__style.heading_col_heading_row_cross,
            column_separator=self.__style.heading_row_cross,
            right_edge=self.__style.heading_row_right_tee,
            filler=self.__style.heading_row_sep,
            previous_content_row=previous_content_row,
            next_content_row=next_content_row,
            top_tee=self.__style.heading_row_top_tee,
            bottom_tee=self.__style.heading_row_bottom_tee,
            heading_col_top_tee=self.__style.heading_col_heading_row_top_tee,
            heading_col_bottom_tee=self.__style.heading_col_heading_row_bottom_tee,
        )

    def __body_to_ascii(self, body: Sequence[Sequence[SupportsStr]]) -> str:
        """Assembles the body of the ascii table

        Returns:
            The body of the ascii table
        """
        # first content row
        output = self.__content_row_to_ascii(body[0]) if len(body) else ""
        for row_index, row in enumerate(body[1:], 1):
            # separator between rows
            output += self.__row_to_ascii(
                left_edge=self.__style.row_left_tee,
                heading_col_sep=self.__style.heading_col_row_cross,
                column_separator=self.__style.col_row_cross,
                right_edge=self.__style.row_right_tee,
                filler=self.__style.row_sep,
                previous_content_row=body[row_index - 1],
                next_content_row=row,
                top_tee=self.__style.col_row_top_tee,
                bottom_tee=self.__style.col_row_bottom_tee,
                heading_col_top_tee=self.__style.heading_col_body_row_top_tee,
                heading_col_bottom_tee=self.__style.heading_col_body_row_bottom_tee,
            )
            # content row
            output += self.__content_row_to_ascii(row)
        return output

    def __str_width(self, text: str) -> int:
        """
        Returns the width of the string in characters for the purposes of monospace
        formatting. This is usually the same as the length of the string, but can be
        different for double-width characters (East Asian Wide and East Asian Fullwidth)
        or zero-width characters (combining characters, zero-width space, etc.)

        Args:
            text: The text to measure

        Returns:
            The width of the string in characters
        """
        width = wcswidth(text) if self.__use_wcwidth else -1
        # if use_wcwidth is False or wcswidth fails, fall back to len
        return width if width >= 0 else len(text)

    @staticmethod
    def __is_number(text: str) -> bool:
        """Returns True if the string is a number, with or without a decimal point"""
        return text.replace(".", "", 1).isdecimal()

    @staticmethod
    def __split_decimal(text: str) -> tuple[str, str]:
        """Splits a string into a tuple of the integer and decimal parts"""
        if "." in text:
            before, after = text.split(".", 1)
            return before, after
        return text, ""

    def to_ascii(self) -> str:
        """Generates a formatted ASCII table

        Returns:
            The generated ASCII table
        """
        # top row of table
        table = self.__top_edge_to_ascii()
        # add table header
        if self.__header:
            table += self.__content_row_to_ascii(self.__header)
            table += self.__heading_sep_to_ascii(
                previous_content_row=self.__header,
                next_content_row=self.__body[0] if self.__body else None,
            )
        # add table body
        if self.__body:
            table += self.__body_to_ascii(self.__body)
        # add table footer
        if self.__footer:
            table += self.__heading_sep_to_ascii(
                previous_content_row=self.__body[-1] if self.__body else None,
                next_content_row=self.__footer,
            )
            table += self.__content_row_to_ascii(self.__footer)
        # bottom row of table
        table += self.__bottom_edge_to_ascii()
        # reurn ascii table
        return table.strip("\n")


def table2ascii(
    header: Sequence[SupportsStr] | None = None,
    body: Sequence[Sequence[SupportsStr]] | None = None,
    footer: Sequence[SupportsStr] | None = None,
    *,
    first_col_heading: bool = False,
    last_col_heading: bool = False,
    column_widths: Sequence[int | None] | None = None,
    alignments: Sequence[Alignment] | Alignment | None = None,
    number_alignments: Sequence[Alignment] | Alignment | None = None,
    cell_padding: int = 1,
    style: TableStyle = PresetStyle.double_thin_compact,
    use_wcwidth: bool = True,
) -> str:
    """Convert a 2D Python table to ASCII text

    Args:
        header (:data:`Optional <typing.Optional>`\ [:class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]]):
            List of column values in the table's header row. All values should be :class:`str`
            or support :class:`str` conversion. If not specified, the table will not have a header row.
        body (:data:`Optional <typing.Optional>`\ [:class:`Sequence <collections.abc.Sequence>`\ [:class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]]]):
            2-dimensional list of values in the table's body. All values should be :class:`str`
            or support :class:`str` conversion. If not specified, the table will not have a body.
        footer (:data:`Optional <typing.Optional>`\ [:class:`Sequence <collections.abc.Sequence>`\ [:class:`SupportsStr`]]):
            List of column values in the table's footer row. All values should be :class:`str`
            or support :class:`str` conversion. If not specified, the table will not have a footer row.
        first_col_heading: Whether to add a header column separator after the first column.
            Defaults to :py:obj:`False`.
        last_col_heading: Whether to add a header column separator before the last column.
            Defaults to :py:obj:`False`.
        column_widths (:data:`Optional <typing.Optional>`\ [:class:`Sequence <collections.abc.Sequence>`\ [:data:`Optional <typing.Optional>`\ [:class:`int`]]]):
            List of widths in characters for each column. Any value of :py:obj:`None`
            indicates that the column width should be determined automatically. If :py:obj:`None`
            is passed instead of a :class:`~collections.abc.Sequence`, all columns will be automatically
            sized. Defaults to :py:obj:`None`.
        alignments: List of alignments for each column
            (ex. [:attr:`Alignment.LEFT`, :attr:`Alignment.CENTER`, :attr:`Alignment.RIGHT`, :attr:`Alignment.DECIMAL`])
            or a single alignment to apply to all columns (ex. :attr:`Alignment.LEFT`).
            If not specified or set to :py:obj:`None`, all columns will be center-aligned.
            Defaults to :py:obj:`None`.

            .. versionchanged:: 1.1.0
                ``alignments`` can now also be specified as a single :class:`Alignment` value to apply to all columns.
        number_alignments: List of alignments for numeric values in each column or a single alignment
            to apply to all columns. This argument can be used to override the alignment of numbers and
            is ignored for non-numeric values. Numeric values include integers, floats, and strings containing only
            :meth:`decimal <str.isdecimal>` characters and at most one decimal point.
            If not specified or set to :py:obj:`None`, numbers will be aligned based on the ``alignments`` argument.
            Defaults to :py:obj:`None`.

            .. versionadded:: 1.1.0
        cell_padding: The minimum number of spaces to add between the cell content and the column
            separator. If set to ``0``, the cell content will be flush against the column separator.
            Defaults to ``1``.
        style: Table style to use for styling (preset styles can be imported).
            Defaults to :ref:`PresetStyle.double_thin_compact`.
        use_wcwidth: Whether to use :func:`wcwidth.wcswidth` to determine the width of each cell instead of
            :func:`len`. The :func:`~wcwidth.wcswidth` function takes into account double-width characters
            (East Asian Wide and East Asian Fullwidth) and zero-width characters (combining characters,
            zero-width space, etc.), whereas :func:`len` determines the width solely based on the number of
            characters in the string. Defaults to :py:obj:`True`.

            .. versionadded:: 1.0.0

    Returns:
        The generated ASCII table
    """
    return TableToAscii(
        header,
        body,
        footer,
        Options(
            first_col_heading=first_col_heading,
            last_col_heading=last_col_heading,
            column_widths=column_widths,
            alignments=alignments,
            number_alignments=number_alignments,
            cell_padding=cell_padding,
            style=style,
            use_wcwidth=use_wcwidth,
        ),
    ).to_ascii()



import sys
from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING or sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class SupportsStr(Protocol):
    """An abstract base class (ABC) with one abstract method :meth:`__str__`"""

    @abstractmethod
    def __str__(self) -> str:
        """Return a string representation of the object"""
        pass



# /usr/bin/env python
import re

from setuptools import setup


def get_name():
    name = ""
    with open("pyproject.toml") as f:
        name = re.search(r'^name = ["\']([^"\']*)["\']', f.read(), re.M)
    if not name:
        raise RuntimeError("name is not set")
    return name.group(1)


def get_version():
    version = ""
    with open("pyproject.toml") as f:
        version = re.search(r'^version = ["\']([^"\']*)["\']', f.read(), re.M)
    if not version:
        raise RuntimeError("version is not set")
    return version.group(1)


def get_dependencies():
    with open("pyproject.toml") as f:
        dependency_match = re.search(r"^dependencies = \[([\s\S]*?)\]", f.read(), re.M)
    if not dependency_match or not dependency_match.group(1):
        return []
    return [
        dependency.strip().strip(",").strip('"')
        for dependency in dependency_match.group(1).split("\n")
        if dependency
    ]


try:
    # check if pyproject.toml can be used to install dependencies and set the version
    setup(
        packages=[get_name()],
        package_data={get_name(): ["py.typed"]},
    )
except Exception:
    # fallback for old versions of pip/setuptools
    setup(
        name=get_name(),
        packages=[get_name()],
        package_data={get_name(): ["py.typed"]},
        version=get_version(),
        install_requires=get_dependencies(),
    )



from .table_style import TableStyle


class PresetStyle:
    """Importable preset styles for more easily selecting a :ref:`TableStyle`.

    See the :ref:`Preset Styles<styles>` for more information on the available styles.

    Example::

        from table2ascii import PresetStyle, table2ascii

        table2ascii(
            header=["Name", "Price", "Category", "Stock"],
            body=[["Milk", "$2.99", "Dairy", 10]],
            style=PresetStyle.ascii_double,  # Use any available preset style
        )

        \"\"\"
        +---------------------------------+
        | Name   Price   Category   Stock |
        +=================================+
        | Milk   $2.99    Dairy      10   |
        +---------------------------------+
        \"\"\"
    """

    thin = TableStyle.from_string("┌─┬─┐││ ├─┼─┤├─┼─┤└┴─┘────┬┴┬┴")
    thin_box = TableStyle.from_string("┌─┬┬┐│││├─┼┼┤├─┼┼┤└┴┴┘┬┴┬┴┬┴┬┴")
    thin_rounded = TableStyle.from_string("╭─┬─╮││ ├─┼─┤├─┼─┤╰┴─╯────┬┴┬┴")
    thin_compact = TableStyle.from_string("┌─┬─┐││ ├─┼─┤     └┴─┘  ──  ┬┴")
    thin_compact_rounded = TableStyle.from_string("╭─┬─╮││ ├─┼─┤     ╰┴─╯  ──  ┬┴")
    thin_thick = TableStyle.from_string("┌─┬─┐││ ┝━┿━┥├─┼─┤└┴─┘──━━┬┴┯┷")
    thin_thick_rounded = TableStyle.from_string("╭─┬─╮││ ┝━┿━┥├─┼─┤╰┴─╯──━━┬┴┯┷")
    thin_double = TableStyle.from_string("┌─┬─┐││ ╞═╪═╡├─┼─┤└┴─┘──══┬┴╤╧")
    thin_double_rounded = TableStyle.from_string("╭─┬─╮││ ╞═╪═╡├─┼─┤╰┴─╯──══┬┴╤╧")
    thick = TableStyle.from_string("┏━┳━┓┃┃ ┣━╋━┫┣━╋━┫┗┻━┛━━━━┳┻┳┻")
    thick_box = TableStyle.from_string("┏━┳┳┓┃┃┃┣━╋╋┫┣━╋╋┫┗┻┻┛┳┻┳┻┳┻┳┻")
    thick_compact = TableStyle.from_string("┏━┳━┓┃┃ ┣━╋━┫     ┗┻━┛  ━━  ┳┻")
    double = TableStyle.from_string("╔═╦═╗║║ ╠═╬═╣╠═╬═╣╚╩═╝════╦╩╦╩")
    double_box = TableStyle.from_string("╔═╦╦╗║║║╠═╬╬╣╠═╬╬╣╚╩╩╝╦╩╦╩╦╩╦╩")
    double_compact = TableStyle.from_string("╔═╦═╗║║ ╠═╬═╣     ╚╩═╝  ══  ╦╩")
    double_thin_box = TableStyle.from_string("╔═╦╤╗║║│╠═╬╪╣╟─╫┼╢╚╩╧╝┬┴╤╧╥╨╦╩")
    double_thin_compact = TableStyle.from_string("╔═╦═╗║║ ╟─╫─╢     ╚╩═╝  ──  ╥╨")
    minimalist = TableStyle.from_string(" ───  │  ━━━  ───  ── ──━━┬┴┯┷")
    borderless = TableStyle.from_string("      ┃  ━              ━━  ━━")
    simple = TableStyle.from_string(" ═    ║  ═              ══  ╦╩")
    ascii = TableStyle.from_string("+-+-+|| +-+-++-+-+++-+----++++")
    ascii_box = TableStyle.from_string("+-+++|||+-++++-+++++++++++++++")
    ascii_compact = TableStyle.from_string("+-+-+|| +-+-+     ++-+  --  --")
    ascii_double = TableStyle.from_string("+-+-+|| +=+=++-+-+++-+--==--==")
    ascii_minimalist = TableStyle.from_string(" ---  |  ===  ---  -- --==--==")
    ascii_borderless = TableStyle.from_string("      |  -              --  --")
    ascii_simple = TableStyle.from_string(" =    |  =              ==  ==")
    ascii_rounded = TableStyle.from_string(r"/===\|| |=|=||-|-|\|=/--==--==")
    ascii_rounded_box = TableStyle.from_string(r"/===\||||=||||-|||\||/--==--==")
    markdown = TableStyle.from_string("     ||||-|||           --  --")
    plain = TableStyle.from_string(" " * 30).set(left_and_right_edge="")
